#define SPHO_ENABLE_DEBUG_PRINT
#include "msph/common.h"
#include "msph/common.h"

#include <sys/queue.h>

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "spho/spho.h"

#include "msph/tree.h"


struct tree_parse {
	struct msph_token_stream	*s;
	struct msph_token		curr;

	int	parser;

	SLIST_HEAD(msph_token_l, msph_token)	head;
};

#define CURR_TOKEN(parse)	(&parse->curr)

#define CTX(parse)		((parse)->s->ctx)

static void	tree_parse_init(struct tree_parse *,
		    struct msph_token_stream *, int);
static int	tree_parse_next_token(struct tree_parse *);

static ssize_t	tree_ind_fprint(FILE *, int, struct msph_tree *);

static void	free_the_tree(struct msph_tree *);

static struct msph_tree_unit	*tree_parse_unit(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_body	*tree_parse_body(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_ident	*tree_parse_ident_extsize(struct tree_parse *,
				    struct msph_tree *, size_t);

static struct msph_tree_namedecl *tree_parse_namedecl(struct tree_parse *,
    struct msph_tree *);
static struct msph_tree_nameref *tree_parse_nameref(struct tree_parse *,
    struct msph_tree *);

static struct msph_tree_dir *tree_parse_tpdef(struct tree_parse *,
				struct msph_tree *);
static struct msph_tree_dir *tree_parse_assert(struct tree_parse *,
				  struct msph_tree *);
static struct msph_tree_dir *tree_parse_nomindecl(
				    struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_dir *tree_parse_membdecl(
				    struct tree_parse *,
				    struct msph_tree *);


static struct msph_tree_tpexpr	*tree_parse_tpexpr(struct tree_parse *,
				    struct msph_tree *);

/* prefix &, |, =>, -> */
static struct msph_tree_tpexpr	*tree_parse_tpexpr_prefix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_conj_prefix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_disj_prefix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_impl_prefix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_arrow_prefix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_subt_prefix(struct tree_parse *,
				    struct msph_tree *);

/* infix &, |, =>, -> */
static struct msph_tree_tpexpr	*tree_parse_tpexpr_infix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_conj_infix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_disj_infix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_impl_infix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_arrow_infix(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_subt_infix(struct tree_parse *,
				    struct msph_tree *);

static struct msph_tree_tpexpr	*tree_parse_tpexpr_atom(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_true(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_false(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_tpname_or_tpappl(
				    struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_tpapplargs(struct tree_parse *,
				    struct msph_tree_nameref*,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_trait(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_box(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_forall(struct tree_parse *,
				    struct msph_tree *);
static struct msph_tree_tpexpr	*tree_parse_paren(struct tree_parse *,
				    struct msph_tree *);

struct msph_tree_root *
msph_tree_makeroot(struct msph_ctx *ctx)
{
	struct msph_tree_root	*root;

	if ((root = calloc(1, sizeof(*root))) == NULL) {
		MSPH_ERR(ctx, MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(root)->type = MSPH_TREE_ROOT;
	TREE(root)->parent = NULL;
	DECOR_INIT(root);

	root->ctx = ctx;

	STAILQ_INIT(&root->head);

	return (root);
}


/*
 * Expect to read next token, otherwise set error and do something
 * readres: (int) result of read
 * parse: struct tree_parse * to perform read
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_READ_NEXT(readres, parse, err_info, on_err) \
	do { \
		if (((readres) = tree_parse_next_token(parse)) == -1) \
			on_err; \
		if ((readres) == 0) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_EOF, \
			    err_info); \
			on_err; \
		} \
	} while (0)

/*
 * Expect current token type, otherwise set error and do something
 * parse: struct tree_parse * to perform read
 * toktype: token type expected
 * err_info: value passed to MSPH_ERR_INFO
 * on_err: statement to perform on error
 */
#define EXPECT_CURR_TOKEN(parse, toktype, err_info, on_err) \
	do { \
		if (CURR_TOKEN(parse)->type != toktype) { \
			MSPH_ERR_INFO(CTX(parse), MSPH_ERR_TREE_NOMATCH, \
			    err_info); \
			on_err; \
		} \
	} while (0)

int
msph_tree_parse(struct msph_token_stream *s, struct msph_tree_root *root,
    int parser)
{
	struct tree_parse p;
	struct msph_tree_unit *unit;

	if (TREE(root)->type != MSPH_TREE_ROOT) {
		MSPH_ERR(s->ctx, MSPH_ERR_INVAL);
		return (-1);
	}

	switch (parser) {
	case MSPH_PARSE_INFIX:
	case MSPH_PARSE_PREFIX:
		tree_parse_init(&p, s, parser);
		break;
	default:
		MSPH_ERR(s->ctx, MSPH_ERR_INVAL);
		return (-1);
	}

	if ((unit = tree_parse_unit(&p, TREE(root))) == NULL)
		return (-1);

	STAILQ_INSERT_TAIL(&root->head, unit, entries);

	return (0);
}

ssize_t
msph_tree_fprint(FILE *f, struct msph_tree *tree) {
	return (tree_ind_fprint(f, 0, tree));
}


static void
tree_parse_init(struct tree_parse *p, struct msph_token_stream *s, int parser)
{
	p->s = s;
	memset(&p->curr, 0, sizeof(p->curr));
	p->curr.type = TOK_NOPE;
	p->parser = parser;
}

static int
tree_parse_next_token(struct tree_parse *p)
{
	int ret;

	SPHO_PRECOND(p->curr.type > TOK_EOF);

	ret = (int)msph_token_stream_read(&p->curr, 1, p->s);

#ifdef SPHO_ENABLE_DEBUG_PRINT
	if (ret == 1) {
		char buf[256];
		msph_token_str(buf, sizeof(buf), CURR_TOKEN(p));
		SPHO_DEBUG_PRINT("tree_parse_read_token: curr=%s\n",
		    buf);
	}
#endif

	if (ret <= 0)
		memset(&p->curr, 0, sizeof(p->curr));
	if (ret < 0)
		p->curr.type = TOK_INVAL;

	return (ret);
}

struct msph_tree_unit *
tree_parse_unit(struct tree_parse *p, struct msph_tree *parent)
{
	size_t namelen;
	struct msph_tree_unit *unit;

	if ((unit = calloc(1, sizeof(*unit))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(unit)->type = MSPH_TREE_UNIT;
	TREE(unit)->parent = parent;
	DECOR_INIT(unit);
	unit->body = NULL;

	if ((namelen = strlcpy(unit->name, p->s->name, sizeof(unit->name)))
	    >= sizeof(unit->name)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TOOLONG);
		goto err;
	}

	tree_parse_next_token(p);

	if ((unit->body = tree_parse_body(p, TREE(unit))) == NULL)
		goto err;

	return (unit);

err:
	free_the_tree(TREE(unit));

	return (NULL);
}

static struct msph_tree_body *
tree_parse_body(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_dir *dir;
	struct msph_tree_body *body;

	if ((body = calloc(1, sizeof(*body))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(body)->type = MSPH_TREE_BODY;
	TREE(body)->parent = parent;
	STAILQ_INIT(&body->head);
	DECOR_INIT(body);

	while (1) {
		dir = NULL;

		switch(CURR_TOKEN(p)->type) {
		case TOK_KW_TYPE:
			dir = tree_parse_tpdef(p, TREE(body));
			break;
		case TOK_KW_NOMINAL:
			dir = tree_parse_nomindecl(p, TREE(body));
			break;
		case TOK_KW_MEMBER:
			dir = tree_parse_membdecl(p, TREE(body));
			break;
		case TOK_KW_ASSERT:
			dir = tree_parse_assert(p, TREE(body));
			break;
		default:
			break;
		}

		if (dir == NULL)
			break;

		STAILQ_INSERT_TAIL(&body->head, dir, entries);
	}

	return (body);
}



struct msph_tree_dir *
tree_parse_tpdef(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpdef *def;
	struct msph_tree_namedecl *name;

	EXPECT_CURR_TOKEN(p, TOK_KW_TYPE, MSPH_TREE_TPDEF, return (NULL));

	if ((def = calloc(1, sizeof(*def))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(def)->type = MSPH_TREE_TPDEF;
	TREE(def)->parent = parent;
	TEXT(def)->pos = CURR_TOKEN(p)->pos;
	STAILQ_INIT(&def->params);
	DECOR_INIT(def);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->name = tree_parse_namedecl(p, TREE(def))) == NULL)
		goto err;

	if (CURR_TOKEN(p)->type == TOK_LBRAK) {
		do {
			EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
			name = tree_parse_namedecl(p, TREE(def));
			if (name == NULL)
				goto err;
			STAILQ_INSERT_TAIL(&def->params, name, entries);

			EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
		} while (CURR_TOKEN(p)->type == TOK_COMMA);

		EXPECT_CURR_TOKEN(p, TOK_RBRAK, MSPH_TREE_TPDEF, goto err);
	}
	EXPECT_CURR_TOKEN(p, TOK_EQUALS, MSPH_TREE_TPDEF, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPDEF, goto err);
	if ((def->tp = tree_parse_tpexpr(p, TREE(def))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)def);
err:
	free_the_tree(TREE(def));
	return (NULL);
}

struct msph_tree_namedecl *
tree_parse_namedecl(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_namedecl *name;

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_NAMEDECL, return (NULL));

	name = (struct msph_tree_namedecl *)tree_parse_ident_extsize(p, parent,
	    sizeof(*name));
	if (name == NULL)
		return (NULL);

	SPHO_ASSERT(TREE(name)->type == MSPH_TREE_IDENT);
	TREE(name)->type = MSPH_TREE_NAMEDECL;

	return (name);
}

struct msph_tree_nameref *
tree_parse_nameref(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_nameref *name;

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_NAMEREF, return (NULL));

	name = (struct msph_tree_nameref *)tree_parse_ident_extsize(p, parent,
	    sizeof(*name));
	if (name == NULL)
		return (NULL);

	SPHO_ASSERT(TREE(name)->type == MSPH_TREE_IDENT);
	TREE(name)->type = MSPH_TREE_NAMEREF;

	return (name);
}

struct msph_tree_ident *
tree_parse_ident_extsize(struct tree_parse *p, struct msph_tree *parent,
    size_t size)
{
	struct msph_tree_ident *id;
	SPHO_PRECOND(size >= sizeof(struct msph_tree_ident));

	EXPECT_CURR_TOKEN(p, TOK_IDENT, MSPH_TREE_IDENT, return (NULL));

	if ((id = calloc(1, size)) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(id)->type = MSPH_TREE_IDENT;
	TREE(id)->parent = parent;
	TEXT(id)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(id);

	if (SPHO_STRLCPY(id->str, CURR_TOKEN(p)->data.str, sizeof(id->str))
	    >= sizeof(id->str)) {
		MSPH_ERR(CTX(p), MSPH_ERR_TREE_TOOLONG);
		goto err;
	}

	tree_parse_next_token(p);

	return (id);
err:
	free_the_tree(TREE(id));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_tpexpr(struct tree_parse *p, struct msph_tree *parent)
{
	switch (p->parser) {
	case MSPH_PARSE_INFIX:
		return (tree_parse_tpexpr_infix(p, parent));
	case MSPH_PARSE_PREFIX:
		return (tree_parse_tpexpr_prefix(p, parent));
	default:
		SPHO_RIP("impossible-invalid-parser");
		return (NULL);
	}

}

struct msph_tree_tpexpr *
tree_parse_tpexpr_prefix(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_tpexpr *tp;

	switch (CURR_TOKEN(p)->type) {
	case TOK_AMP:
		tp = tree_parse_conj_prefix(p, parent);
		break;
	case TOK_PIPE:
		tp = tree_parse_disj_prefix(p, parent);
		break;
	case TOK_IMPL:
		tp = tree_parse_impl_prefix(p, parent);
		break;
	case TOK_SUB:
		tp = tree_parse_subt_prefix(p, parent);
		break;
	case TOK_RARROW:
		tp = tree_parse_arrow_prefix(p, parent);
		break;
	case TOK_KW_BOX:
		tp = tree_parse_box(p, parent);
		break;
	case TOK_KW_FORALL:
		tp = tree_parse_forall(p, parent);
		break;
	case TOK_IDENT:
		tp = tree_parse_tpname_or_tpappl(p, parent);
		break;
	case TOK_CONST_TRUE:
		tp = tree_parse_true(p, parent);
		break;
	case TOK_CONST_FALSE:
		tp = tree_parse_false(p, parent);
		break;
	case TOK_LPAREN:
		tp = tree_parse_paren(p, parent);
		break;
	case TOK_LBRACE:
		tp = tree_parse_trait(p, parent);
		break;
	default:
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_NOMATCH, MSPH_TREE_TPEXPR);
		tp = NULL;
	}

	SPHO_DEBUG_PRINT("returning tp type=%x\n", TREE(tp)->type);
	return (tp);
}

struct msph_tree_tpexpr *
tree_parse_conj_prefix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_conj *conj;

	EXPECT_CURR_TOKEN(p, TOK_AMP, MSPH_TREE_CONJ, return (NULL));

	if ((conj = calloc(1, sizeof(*conj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(conj)->type = MSPH_TREE_CONJ;
	TREE(conj)->parent = parent;
	TEXT(conj)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(conj);
	conj->ltp = NULL;
	conj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);
	if ((conj->ltp = tree_parse_tpexpr_prefix(p, TREE(conj))) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);
	if ((conj->rtp = tree_parse_tpexpr_prefix(p, TREE(conj)))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)conj);
err:
	free_the_tree((struct msph_tree *)conj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_disj_prefix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_disj *disj;

	EXPECT_CURR_TOKEN(p, TOK_PIPE, MSPH_TREE_DISJ, return (NULL));

	if ((disj = calloc(1, sizeof(*disj))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(disj)->type = MSPH_TREE_DISJ;
	TREE(disj)->parent = parent;
	TEXT(disj)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(disj);
	disj->ltp = NULL;
	disj->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);
	if ((disj->ltp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)disj)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);
	if ((disj->rtp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)disj))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)disj);
err:
	free_the_tree((struct msph_tree *)disj);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_impl_prefix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_impl *impl;

	EXPECT_CURR_TOKEN(p, TOK_IMPL, MSPH_TREE_IMPL, return (NULL));

	if ((impl = calloc(1, sizeof(*impl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(impl)->type = MSPH_TREE_IMPL;
	TREE(impl)->parent = parent;
	TEXT(impl)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(impl);
	impl->ltp = NULL;
	impl->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);
	if ((impl->ltp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)impl)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);
	if ((impl->rtp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)impl))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)impl);
err:
	free_the_tree((struct msph_tree *)impl);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_subt_prefix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_subt *subt;

	EXPECT_CURR_TOKEN(p, TOK_SUB, MSPH_TREE_SUBT, return (NULL));

	if ((subt = calloc(1, sizeof(*subt))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(subt)->type = MSPH_TREE_SUBT;
	TREE(subt)->parent = parent;
	TEXT(subt)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(subt);
	subt->ltp = NULL;
	subt->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_SUBT, goto err);
	if ((subt->ltp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)subt)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_SUBT, goto err);
	if ((subt->rtp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)subt))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)subt);
err:
	free_the_tree((struct msph_tree *)subt);

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_arrow_prefix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_arrow *arrow;

	EXPECT_CURR_TOKEN(p, TOK_RARROW, MSPH_TREE_ARROW, return (NULL));

	if ((arrow = calloc(1, sizeof(*arrow))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(arrow)->type = MSPH_TREE_ARROW;
	TREE(arrow)->parent = parent;
	TEXT(arrow)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(arrow);
	arrow->ltp = NULL;
	arrow->rtp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);
	if ((arrow->ltp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)arrow)) 
	    == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);
	if ((arrow->rtp = tree_parse_tpexpr_prefix(p, (struct msph_tree *)arrow))
	    == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)arrow);
err:
	free_the_tree((struct msph_tree *)arrow);

	return (NULL);
}

static struct msph_tree_tpexpr *
tree_parse_tpexpr_infix(struct tree_parse *p, struct msph_tree *parent)
{
	return (tree_parse_impl_infix(p, parent));
}

/*
 * infix type algebraics parsing.
 *
 * In reverse order of precedence (later means higher precedence):
 *  [op    _ associativity]
 * - impl  _ right
 * - subt  - right
 * - disj  _ left (both)
 * - conj  _ left (both)
 * - arrow _ right
 *
 * 
 */
static struct msph_tree_tpexpr *
tree_parse_impl_infix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpexpr *left, *right;
	struct msph_tree_impl *impl;
	struct msph_text_pos infix_pos;

	impl = NULL;
	left = right = NULL;

	if ((left = tree_parse_subt_infix(p, NULL)) == NULL) {
		return (NULL);
	}

	if (CURR_TOKEN(p)->type != TOK_IMPL) {
		TREE(left)->parent = parent;
		return (left);
	}
	infix_pos = CURR_TOKEN(p)->pos;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_IMPL, goto err);

	if ((right = tree_parse_impl_infix(p, NULL)) == NULL)
		goto err;

	if ((impl = calloc(1, sizeof(*impl))) ==  NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}

	TREE(impl)->type = MSPH_TREE_IMPL;
	TREE(impl)->parent = parent;
	TEXT(impl)->pos = infix_pos;
	DECOR_INIT(impl);

	impl->ltp = left;
	TREE(left)->parent = TREE(impl);

	impl->rtp = right;
	TREE(right)->parent = TREE(impl);

	return ((struct msph_tree_tpexpr *)impl);
err:
	if (left != NULL)
		free_the_tree(TREE(left));
	if (right != NULL)
		free_the_tree(TREE(right));
	return (NULL);
}

static struct msph_tree_tpexpr *
tree_parse_subt_infix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpexpr *left, *right;
	struct msph_tree_subt *subt;
	struct msph_text_pos infix_pos;

	subt = NULL;
	left = right = NULL;

	if ((left = tree_parse_disj_infix(p, NULL)) == NULL) {
		return (NULL);
	}

	if (CURR_TOKEN(p)->type != TOK_SUB) {
		TREE(left)->parent = parent;
		return (left);
	}
	infix_pos = CURR_TOKEN(p)->pos;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_SUBT, goto err);

	if ((right = tree_parse_subt_infix(p, NULL)) == NULL)
		goto err;

	if ((subt = calloc(1, sizeof(*subt))) ==  NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}

	TREE(subt)->type = MSPH_TREE_SUBT;
	TREE(subt)->parent = parent;
	TEXT(subt)->pos = infix_pos;
	DECOR_INIT(subt);

	subt->ltp = left;
	TREE(left)->parent = TREE(subt);

	subt->rtp = right;
	TREE(right)->parent = TREE(subt);

	return ((struct msph_tree_tpexpr *)subt);
err:
	if (left != NULL)
		free_the_tree(TREE(left));
	if (right != NULL)
		free_the_tree(TREE(right));
	return (NULL);
}

static struct msph_tree_tpexpr *
tree_parse_disj_infix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpexpr *left, *right;
	struct msph_tree_disj *disj;
	struct msph_text_pos infix_pos;

	left = right = NULL;
	disj = NULL;

	if ((left = tree_parse_conj_infix(p, NULL)) == NULL)
		return (NULL);

	TREE(left)->parent = parent;

	while (CURR_TOKEN(p)->type == TOK_PIPE) {
		infix_pos = CURR_TOKEN(p)->pos;
		EXPECT_READ_NEXT(res, p, MSPH_TREE_DISJ, goto err);

		if ((right = tree_parse_conj_infix(p, NULL)) == NULL)
			goto err;

		if ((disj = calloc(1, sizeof(*disj))) == NULL) {
			MSPH_ERR(CTX(p), MSPH_ERR_SYS);
			goto err;
		}

		TREE(disj)->type = MSPH_TREE_DISJ;
		TREE(disj)->parent = parent;
		TEXT(disj)->pos = infix_pos;
		DECOR_INIT(disj);

		disj->ltp = left;
		TREE(left)->parent = TREE(disj);
		disj->rtp = right;
		TREE(right)->parent = TREE(disj);

		left = (struct msph_tree_tpexpr *)disj;
		right = NULL;
	}

	SPHO_POSTCOND(TREE(left)->parent == parent);

	return (left);
err:
	if (right != NULL)
		free_the_tree(TREE(right));

	free_the_tree(TREE(left));
	return (NULL);
}

static struct msph_tree_tpexpr *
tree_parse_conj_infix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpexpr *left, *right;
	struct msph_tree_conj *conj;
	struct msph_text_pos infix_pos;

	left = right = NULL;
	conj = NULL;

	if ((left = tree_parse_arrow_infix(p, NULL)) == NULL)
		return (NULL);

	TREE(left)->parent = parent;

	while (CURR_TOKEN(p)->type == TOK_AMP) {
		infix_pos = CURR_TOKEN(p)->pos;
		EXPECT_READ_NEXT(res, p, MSPH_TREE_CONJ, goto err);

		if ((right = tree_parse_arrow_infix(p, NULL)) == NULL)
			goto err;

		if ((conj = calloc(1, sizeof(*conj))) == NULL) {
			MSPH_ERR(CTX(p), MSPH_ERR_SYS);
			goto err;
		}

		TREE(conj)->type = MSPH_TREE_CONJ;
		TREE(conj)->parent = parent;
		TEXT(conj)->pos = infix_pos;
		DECOR_INIT(conj);

		conj->ltp = left;
		TREE(left)->parent = TREE(conj);
		conj->rtp = right;
		TREE(right)->parent = TREE(conj);

		left = (struct msph_tree_tpexpr *)conj;
		right = NULL;
	}

	SPHO_POSTCOND(TREE(left)->parent == parent);

	return (left);
err:
	if (right != NULL)
		free_the_tree(TREE(right));

	free_the_tree(TREE(left));
	return (NULL);
}
static struct msph_tree_tpexpr *
tree_parse_arrow_infix(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpexpr *left, *right;
	struct msph_tree_arrow *arrow;
	struct msph_text_pos infix_pos;

	arrow = NULL;
	left = right = NULL;

	if ((left = tree_parse_tpexpr_atom(p, NULL)) == NULL) {
		return (NULL);
	}

	if (CURR_TOKEN(p)->type != TOK_RARROW) {
		TREE(left)->parent = parent;
		return (left);
	}
	infix_pos = CURR_TOKEN(p)->pos;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ARROW, goto err);

	if ((right = tree_parse_arrow_infix(p, NULL)) == NULL)
		goto err;

	if ((arrow = calloc(1, sizeof(*arrow))) ==  NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}

	TREE(arrow)->type = MSPH_TREE_ARROW;
	TREE(arrow)->parent = parent;
	TEXT(arrow)->pos = infix_pos;
	DECOR_INIT(arrow);

	arrow->ltp = left;
	TREE(left)->parent = TREE(arrow);

	arrow->rtp = right;
	TREE(right)->parent = TREE(arrow);

	return ((struct msph_tree_tpexpr *)arrow);
err:
	if (left != NULL)
		free_the_tree(TREE(left));
	if (right != NULL)
		free_the_tree(TREE(right));
	return (NULL);
}


struct msph_tree_tpexpr *
tree_parse_tpexpr_atom(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_tpexpr *tp;

	switch (CURR_TOKEN(p)->type) {
	case TOK_KW_BOX:
		tp = tree_parse_box(p, parent);
		break;
	case TOK_KW_FORALL:
		tp = tree_parse_forall(p, parent);
		break;
	case TOK_IDENT:
		tp = tree_parse_tpname_or_tpappl(p, parent);
		break;
	case TOK_CONST_TRUE:
		tp = tree_parse_true(p, parent);
		break;
	case TOK_CONST_FALSE:
		tp = tree_parse_false(p, parent);
		break;
	case TOK_LPAREN:
		tp = tree_parse_paren(p, parent);
		break;
	case TOK_LBRACE:
		tp = tree_parse_trait(p, parent);
		break;
	default:
		MSPH_ERR_INFO(CTX(p), MSPH_ERR_TREE_NOMATCH, MSPH_TREE_TPEXPR);
		tp = NULL;
	}

	SPHO_DEBUG_PRINT("returning tp type=%x\n", TREE(tp)->type);
	return (tp);
}
struct msph_tree_tpexpr *
tree_parse_box(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_box *box;

	EXPECT_CURR_TOKEN(p, TOK_KW_BOX, MSPH_TREE_BOX, return (NULL));

	if ((box = calloc(1, sizeof(*box))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(box)->type = MSPH_TREE_BOX;
	TREE(box)->parent = parent;
	TEXT(box)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(box);
	box->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_BOX, goto err);
	if ((box->inner = tree_parse_tpexpr(p, TREE(box))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)box);
err:
	free_the_tree(TREE(box));
	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_forall(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_forall *fa;

	EXPECT_CURR_TOKEN(p, TOK_KW_FORALL, MSPH_TREE_FORALL, return (NULL));

	if ((fa = calloc(1, sizeof(*fa))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(fa)->type = MSPH_TREE_FORALL;
	TREE(fa)->parent = parent;
	TEXT(fa)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(fa);
	fa->name = NULL;
	fa->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->name = tree_parse_namedecl(p, TREE(fa))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_DOT, MSPH_TREE_FORALL, goto err);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_FORALL, goto err);
	if ((fa->inner = tree_parse_tpexpr(p, TREE(fa))) == NULL)
		goto err;

	return ((struct msph_tree_tpexpr *)fa);
err:
	free_the_tree(TREE(fa));
	return (NULL);
}

/* parse name or appl */
struct msph_tree_tpexpr *
tree_parse_tpname_or_tpappl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpname *tpname;
	struct msph_tree_tpexpr *ret;
	struct msph_tree_nameref *nameref;

	ret = NULL;
	nameref = NULL;

	if ((nameref = tree_parse_nameref(p, NULL)) == NULL)
		return (NULL);

	if (CURR_TOKEN(p)->type == TOK_LBRAK) {
		EXPECT_READ_NEXT(res, p, MSPH_TREE_TPAPPL, goto err);
		if ((ret = tree_parse_tpapplargs(p, nameref, parent)) == NULL) 
			goto err;
		return (ret);
	}

	if ((tpname = calloc(1, sizeof(*tpname))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		goto err;
	}
	TREE(tpname)->type = MSPH_TREE_TPNAME;
	TREE(tpname)->parent = parent;

	TEXT(tpname)->pos = TEXT(nameref)->pos;
	DECOR_INIT(tpname);
	TREE(nameref)->parent = TREE(tpname);
	tpname->name = nameref;

	return ((struct msph_tree_tpexpr *)tpname);
err:
	free_the_tree(TREE(nameref));

	return (NULL);
}

/**
 * Parse argument list part [...], of type application A[...].
 *
 * @param p Parse state
 * @param id Parsed ident A (on failure, this is not freed)
 * @param parent Parent of resulting appl node
 *
 * @return tpexpr node representing A[...].
 */
struct msph_tree_tpexpr *
tree_parse_tpapplargs(struct tree_parse *p, struct msph_tree_nameref *name,
    struct msph_tree *parent)
{
	int res;
	struct msph_tree_tpappl *tpappl;
	struct msph_tree_tpexpr *tp;

	tp = NULL;

	EXPECT_CURR_TOKEN(p, TOK_LBRAK, MSPH_TREE_TPAPPL, return (NULL));

	if ((tpappl = calloc(1, sizeof(*tpappl))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(tpappl)->type = MSPH_TREE_TPAPPL;
	TREE(tpappl)->parent = parent;
	DECOR_INIT(tpappl);
	tpappl->name = NULL;
	STAILQ_INIT(&tpappl->args);

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TPAPPL, goto err);

	if ((tp = tree_parse_tpexpr(p, TREE(tpappl))) == NULL)
		goto err;

	/* more arguments */
	while (CURR_TOKEN(p)->type == TOK_COMMA) {

		EXPECT_READ_NEXT(res, p, MSPH_TREE_TPAPPL, goto err);
		if ((tp = tree_parse_tpexpr(p, TREE(tpappl))) == NULL)
			goto err;

		STAILQ_INSERT_TAIL(&tpappl->args, tp, entries);
	}

	EXPECT_CURR_TOKEN(p, TOK_RBRAK, MSPH_TREE_TPAPPL, goto err);

	/* finally, appl takes ownership of id */
	TREE(name)->parent = TREE(tpappl);
	tpappl->name = name;
	TEXT(tpappl)->pos = TEXT(name)->pos;

	tree_parse_next_token(p);

	return ((struct msph_tree_tpexpr *)tpappl);
err:
	free_the_tree(TREE(tpappl));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_true(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *t;

	EXPECT_CURR_TOKEN(p, TOK_CONST_TRUE, MSPH_TREE_TRUE, return (NULL));

	if ((t = calloc(1, sizeof(*t))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(t)->type = MSPH_TREE_TRUE;
	TREE(t)->parent = parent;
	TEXT(t)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(t);

	tree_parse_next_token(p);

	return ((struct msph_tree_tpexpr *)t);
}

struct msph_tree_tpexpr *
tree_parse_false(struct tree_parse *p, struct msph_tree *parent)
{
	struct msph_tree_true *f;

	EXPECT_CURR_TOKEN(p, TOK_CONST_FALSE, MSPH_TREE_FALSE, return (NULL));

	if ((f = calloc(1, sizeof(*f))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(f)->type = MSPH_TREE_FALSE;
	TREE(f)->parent = parent;
	TEXT(f)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(f);

	tree_parse_next_token(p);

	return ((struct msph_tree_tpexpr *)f);
}

struct msph_tree_tpexpr *
tree_parse_paren(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_paren *par;

	EXPECT_CURR_TOKEN(p, TOK_LPAREN, MSPH_TREE_FALSE, return (NULL));

	if ((par = calloc(1, sizeof(*par))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(par)->type = MSPH_TREE_PAREN;
	TREE(par)->parent = parent;
	TEXT(par)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(par);
	par->inner = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_PAREN, goto err);

	if ((par->inner = tree_parse_tpexpr(p, TREE(par))) == NULL)
		goto err;

	EXPECT_CURR_TOKEN(p, TOK_RPAREN, MSPH_TREE_PAREN, goto err);

	tree_parse_next_token(p);

	return ((struct msph_tree_tpexpr *)par);
err:
	free_the_tree(TREE(par));

	return (NULL);
}

struct msph_tree_tpexpr *
tree_parse_trait(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_trait *trait;

	EXPECT_CURR_TOKEN(p, TOK_LBRACE, MSPH_TREE_TRAIT, return (NULL));

	if ((trait = calloc(1, sizeof(*trait))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(trait)->type = MSPH_TREE_TRAIT;
	TREE(trait)->parent = parent;
	TEXT(trait)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(trait);
	trait->body = NULL;

	if ((trait->body = tree_parse_body(p, TREE(trait))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_TRAIT, goto err);
	EXPECT_CURR_TOKEN(p, TOK_RBRACE, MSPH_TREE_TRAIT, goto err);

	return ((struct msph_tree_tpexpr *)trait);
err:
	free_the_tree(TREE(trait));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_nomindecl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_namedecl *name;
	struct msph_tree_nomindecl *nomind;

	EXPECT_CURR_TOKEN(p, TOK_KW_NOMINAL, MSPH_TREE_NOMINDECL,
	    return (NULL));

	if ((nomind = calloc(1, sizeof(*nomind))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(nomind)->type = MSPH_TREE_NOMINDECL;
	TREE(nomind)->parent = parent;
	TEXT(nomind)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(nomind);
	nomind->name = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_NOMINDECL, goto err);
	if ((name = tree_parse_namedecl(p, TREE(nomind))) == NULL)
		goto err;

	nomind->name = name;

	return ((struct msph_tree_dir *)nomind);
err:
	free_the_tree(TREE(nomind));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_membdecl(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_membdecl *membd;

	EXPECT_CURR_TOKEN(p, TOK_KW_MEMBER, MSPH_TREE_MEMBDECL,
	    return (NULL));

	if ((membd = calloc(1, sizeof(*membd))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(membd)->type = MSPH_TREE_MEMBDECL;
	TREE(membd)->parent = parent;
	TEXT(membd)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(membd);
	membd->name = NULL;
	membd->tp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);
	if ((membd->name = tree_parse_namedecl(p, TREE(membd))) == NULL)
		goto err;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);
	EXPECT_CURR_TOKEN(p, TOK_COLON, MSPH_TREE_MEMBDECL,
	    goto err);
	EXPECT_READ_NEXT(res, p, MSPH_TREE_MEMBDECL, goto err);

	if ((membd->tp = tree_parse_tpexpr(p, TREE(membd))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)membd);
err:
	free_the_tree(TREE(membd));

	return (NULL);
}

struct msph_tree_dir *
tree_parse_assert(struct tree_parse *p, struct msph_tree *parent)
{
	int res;
	struct msph_tree_assert *ass;

	EXPECT_CURR_TOKEN(p, TOK_KW_ASSERT, MSPH_TREE_ASSERT, return (NULL));

	if ((ass = calloc(1, sizeof(*ass))) == NULL) {
		MSPH_ERR(CTX(p), MSPH_ERR_SYS);
		return (NULL);
	}
	TREE(ass)->type = MSPH_TREE_ASSERT;
	TREE(ass)->parent = parent;
	TEXT(ass)->pos = CURR_TOKEN(p)->pos;
	DECOR_INIT(ass);
	ass->tp = NULL;

	EXPECT_READ_NEXT(res, p, MSPH_TREE_ASSERT, goto err);
	if ((ass->tp = tree_parse_tpexpr(p, TREE(ass))) == NULL)
		goto err;

	return ((struct msph_tree_dir *)ass);
err:
	free_the_tree(TREE(ass));

	return (NULL);
}

static void
free_the_tree(struct msph_tree *tree)
{
	/* will anyone ever free the tree? :´( */
	SPHO_POSTCOND((void *)tree != (void *)free);
	// SPHO_POSTCOND(0);
}

#define MSPH_TREE_INDENT	"    "

__attribute__((format(printf, 3, 4)))
static ssize_t
indent_fprintf(FILE *f, int indent, const char *fmt, ...)
{
	int res;
	ssize_t ret;
	int i;
	va_list ap;

	ret = 0;
	va_start(ap, fmt);

	for (i = 0; i < indent; i++) {
		res = fprintf(f, "%s", MSPH_TREE_INDENT);
		if (res < 0) {
			return ((ssize_t)res);
		}
		ret += (ssize_t)res;
	}

	res = vfprintf(f, fmt, ap);
	if (res < 0) {
		return ((ssize_t)res);
	}
	ret += (ssize_t)res;
	va_end(ap);

	return (ret);
}

static ssize_t
tree_ind_fprint(FILE *f, int indent, struct msph_tree *tree)
{
	ssize_t ret, res;
	struct msph_tree_root *root;
	struct msph_tree_unit *unit;
	struct msph_tree_body *body;
	struct msph_tree_dir *dir;
	struct msph_tree_nomindecl *nomd;
	struct msph_tree_assert *ass;
	struct msph_tree_tpdef *tpdef;
	struct msph_tree_membdecl *membdecl;
	struct msph_tree_conj *conj;
	struct msph_tree_disj *disj;
	struct msph_tree_impl *impl;
	struct msph_tree_subt *subt;
	struct msph_tree_arrow *arrow;
	struct msph_tree_box *box;
	struct msph_tree_forall *forall;
	struct msph_tree_tpappl *tpappl;
	struct msph_tree_trait *trait;
	struct msph_tree_tpname *name;
	struct msph_tree_paren *paren;
	struct msph_tree_ident *ident;
	struct msph_tree_tpexpr *tp;
	const char *ident_kind;

	ret = 0;
	switch (TREE_ID(tree)) {
	case MSPH_TREE_ROOT:
		root = (struct msph_tree_root *)tree;
		if ((res = indent_fprintf(f, indent, "(root\n")) < 0)
			return (res);
		STAILQ_FOREACH(unit, &root->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, TREE(unit)))
			    < 0) {
				return (res);
			}
			ret += res;
		}
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		break;
	case MSPH_TREE_UNIT:
		unit = (struct msph_tree_unit *)tree;
		if ((res = indent_fprintf(f, indent, "(unit:%s\n", unit->name))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(unit->body))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_BODY:
		body = (struct msph_tree_body *)tree;
		if ((res = indent_fprintf(f, indent, "(body\n")) < 0)
			return (res);
		ret += res;
		STAILQ_FOREACH(dir, &body->head, entries) {
			if ((res = tree_ind_fprint(f, indent+1, TREE(dir)))
			    < 0) {
				return (res);
			}
			ret += res;
		}
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_NOMINDECL:
		nomd = (struct msph_tree_nomindecl *)tree;
		if ((res = indent_fprintf(f, indent, "(nomindecl:%u,%u\n",
		    TEXT(nomd)->pos.line, TEXT(nomd)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(nomd->name))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_ASSERT:
		ass = (struct msph_tree_assert *)tree;
		if ((res = indent_fprintf(f, indent, "(assert:%u,%u\n",
		    TEXT(ass)->pos.line, TEXT(ass)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(ass->tp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPDEF:
		tpdef = (struct msph_tree_tpdef *)tree;
		if ((res = indent_fprintf(f, indent, "(tpdef:%u,%u\n",
		    TEXT(tpdef)->pos.line, TEXT(tpdef)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(tpdef->name)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(tpdef->tp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_MEMBDECL:
		membdecl = (struct msph_tree_membdecl *)tree;
		if ((res = indent_fprintf(f, indent, "(membdecl:%u,%u\n",
		    TEXT(membdecl)->pos.line, TEXT(membdecl)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(membdecl->name)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(membdecl->tp)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_CONJ:
		conj = (struct msph_tree_conj *)tree;
		if ((res = indent_fprintf(f, indent, "(&conj:%u,%u\n",
		    TEXT(conj)->pos.line, TEXT(conj)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(conj->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(conj->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_DISJ:
		disj = (struct msph_tree_disj *)tree;
		if ((res = indent_fprintf(f, indent, "(|disj:%u,%u\n",
		    TEXT(disj)->pos.line, TEXT(disj)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(disj->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(disj->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_IMPL:
		impl = (struct msph_tree_impl *)tree;
		if ((res = indent_fprintf(f, indent, "(=>impl:%u,%u\n",
		    TEXT(impl)->pos.line, TEXT(impl)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(impl->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(impl->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_SUBT:
		subt = (struct msph_tree_subt *)tree;
		if ((res = indent_fprintf(f, indent, "(<:subt:%u,%u\n",
		    TEXT(subt)->pos.line, TEXT(subt)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(subt->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(subt->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_ARROW:
		arrow = (struct msph_tree_arrow *)tree;
		if ((res = indent_fprintf(f, indent, "(->arrow:%u,%u\n",
		    TEXT(arrow)->pos.line, TEXT(arrow)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(arrow->ltp))) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(arrow->rtp))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_BOX:
		box = (struct msph_tree_box *)tree;
		if ((res = indent_fprintf(f, indent, "(box:%u,%u\n",
		    TEXT(box)->pos.line, TEXT(box)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(box->inner))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_FORALL:
		forall = (struct msph_tree_forall *)tree;
		if ((res = indent_fprintf(f, indent, "(forall:%u,%u\n",
		    TEXT(forall)->pos.line, TEXT(forall)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(forall->name)))
		    < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(forall->inner)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_PAREN:
		paren = (struct msph_tree_paren *)tree;
		if ((res = indent_fprintf(f, indent, "(paren:%u,%u\n",
		    TEXT(paren)->pos.line, TEXT(paren)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(paren->inner)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPAPPL:
		tpappl = (struct msph_tree_tpappl *)tree;
		if ((res = indent_fprintf(f, indent, "(tpappl:%u,%u\n",
		    TEXT(tpappl)->pos.line, TEXT(tpappl)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(tpappl->name)))
		    < 0) {
			return (res);
		}
		ret += res;
		STAILQ_FOREACH(tp, &tpappl->args, entries) {
			if ((res = tree_ind_fprint(f, indent+1, TREE(tp)))
			    < 0) {
				return (res);
			}
			ret += res;
		}
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TPNAME:
		name = (struct msph_tree_tpname *)tree;
		if ((res = indent_fprintf(f, indent, "(name:%u,%u\n",
		    TEXT(name)->pos.line, TEXT(name)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(name->name))) < 0)
			return (res);
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_TRAIT:
		trait = (struct msph_tree_trait *)tree;
		if ((res = indent_fprintf(f, indent, "(trait:%u,%u\n",
		    TEXT(trait)->pos.line, TEXT(trait)->pos.col)) < 0)
			return (res);
		ret += res;
		if ((res = tree_ind_fprint(f, indent+1, TREE(trait->body)))
		    < 0) {
			return (res);
		}
		ret += res;
		if ((res = indent_fprintf(f, indent, ")\n")) < 0)
			return (res);
		ret += res;
		break;
	case MSPH_TREE_NAMEDECL:
	case MSPH_TREE_NAMEREF:
	case MSPH_TREE_IDENT:
		switch (TREE_ID(tree)) {
		case MSPH_TREE_NAMEDECL:
			ident_kind = "/namedecl";
			break;
		case MSPH_TREE_NAMEREF:
			ident_kind = "/nameref";
			break;
		default:
			ident_kind = "";
		}
		ident = IDENT(tree);
		if ((res = indent_fprintf(f, indent, "(ident%s:%u,%u:%s)\n",
		    ident_kind, TEXT(ident)->pos.line, TEXT(ident)->pos.col,
		    ident->str)) < 0) {
			return (res);
		}
		ret += res;
		break;
	case MSPH_TREE_TRUE:
		if ((res = indent_fprintf(f, indent, "(true:%u,%u)\n",
		    TEXT(tree)->pos.line, TEXT(tree)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		break;
	case MSPH_TREE_FALSE:
		if ((res = indent_fprintf(f, indent, "(false:%u,%u)\n",
		    TEXT(tree)->pos.line, TEXT(tree)->pos.col)) < 0) {
			return (res);
		}
		ret += res;
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}
